#!/usr/bin/perl

sub calc_time {
    local ($fhh, $fmm, $fss, $hh, $mm, $ss) = @_ ; 
    local ($diff) ; 

    if ($hh < $fhh) { # a day change occured
	$diff = 24 ; 
    }
    else {
	$diff = 0 ; 
    }

    $diff = $diff + ($hh - $fhh) ;
    $diff = $diff*60 ; 
    $diff = $diff + ($mm - $fmm) ; 
    $diff = $diff*60 ; 
    $diff = $diff + ($ss - $fss); 

    return $diff ; 
}

sub translate {
    local ($line) = @_;

    $line =~ s/128\.119\.41\.224/umass/g;
    $line =~ s/128\.119\.241\.136/umass/g;
    $line =~ s/134\.155\.48\.96/unima/g;
    $line =~ s/128\.16\.64\.73/ucl/g;
    $line =~ s/208\.219\.91\.202/nokia/g;
    $line =~ s/24\.6\.255\.61/cable-modem/g;
    $line =~ s/192\.150\.187\.31/aciri\(lemming\)/g ;
    $line =~ s/192\.150\.187\.61/aciri\(hyrax\)/g ;
    $line =~ s/192\.150\.187\.23/aciri\(ox\)/g ;
    $line =~ s/192\.150\.187\.20/aciri\(aak\)/g ;
    $line =~ s/192\.150\.187\.12/aciri\(griffin\)/g ;
    return $line;
}

if (@ARGV < 2) {
    printf "usage parse.pl <filename> <ports>\n";
    exit;
}

$ignorefirstseconds = 60; # affects avg throughput of the flows
$secondsperinterval = 1; # number of seconds per measurement interval (granularity)
$smoothing = 1;           # average samples over $n$ seconds for smoothing at low bitrates

@sm_array = ();
for ($i = 0; $i < $smoothing; ++$i) {
    push @sm_array, 0;
}

$firstline = 1;
$max_time = 0;
$udp_nr = 0;
@udp_data = ();
$udp_sum = ();
$tcp_nr = 0;
@tcp_data = ();
$tcp_sum = ();
$udp_ft = ();
$udp_lt = ();
$tcp_ft = ();
$tcp_lt = ();

$ARGV[0] =~ /\.(\d+)$/;
$cur_nr = $1;
$pref = "0" if ($cur_nr < 100);
$pref = "00" if ($cur_nr < 10);

open(fp, $ARGV[0]) or die "can't open log file $ARGV[0]\n";
@log = <fp>;
close(fp);

printf "analyzing %s\n", $ARGV[0];
for ($port = 1; $port < @ARGV; $port++) {
    printf "port %s\n", $ARGV[$port];
    $prev_time = 0;
    $data_sec = 0;
    $data_pkts = 0;
    $lost_pkts = 0;
    $high_ack = 0;
    $maxseq = -1;
    $lastseq = -1;

    foreach $line (@log) {
	
	($hh,$mm,$ss,$from,$to,$flag) = split(/[\s>:\(\)]+/,$line); 
	if ($flag =~ /\.|P/) {
	    # reparse TCP
	    ($hh, $mm, $ss, $from, $to, $flag, $start, $end, $size) = 
		split (/[\s>:\(\)]+/, $line);
	} else {
	    if ($flag =~ /UDP/i) {
		# reparse UDP
		($hh, $mm, $ss, $from, $to, $flag, $size) = 
		    split (/[\s>:\(\)]+/, $line); 
	    } else {
		# skip SYNC etc.
		next;
	    }
	}
	#
	# need to check for firstline to convert absolute time 
	# into relative
	#
	if ($firstline == 1) {
	    $fhh = $hh ;
	    $fmm = $mm ; 
	    $fss = $ss ; 
	    $firstline = 0;
	}

	$time = int(calc_time($fhh, $fmm, $fss, $hh, $mm, $ss));

	# check for ACK
	#if (($from =~ /$ARGV[$port]/) && ($flag !~ /udp/i) && ($start =~ /ack/i)) {
	#    if ($end <= $high_ack) {
	#        printf "%7.3f: ack %i, %i\n", calc_time($fhh, $fmm, $fss, $hh, $mm, $ss), $end, $high_ack;
	#    }
	#    $high_ack = $end;
	#    next;
	#}

	# check for data packet
	if ($to =~ /$ARGV[$port]/) {
	    if ($time - $prev_time >= $secondsperinterval) {
		if ($prev_time == 0) {
		    if ($flag =~ /udp/i) {
			$udp_nr++;
			$fname = ">udp$udp_nr.log.$cur_nr";
			$udp_data[$udp_nr] = 0;
			if ($time > $ignorefirstseconds) {
			    $udp_ft[$udp_nr] = $time;
			} else {
			    $udp_ft[$udp_nr] = $ignorefirstseconds;
			}
			$udp_lt[$udp_nr] = 0;
		    } else {
			$tcp_nr++;
			$fname = ">tcp$tcp_nr.log.$cur_nr";
			$tcp_data[$tcp_nr] = 0;
			if ($time > $ignorefirstseconds) {
			    $tcp_ft[$tcp_nr] = $time;
			} else {
			    $tcp_ft[$tcp_nr] = $ignorefirstseconds;
			}
			$tcp_lt[$tcp_nr] = 0;
		    }
		    open(outfile, $fname) or die "write error\n";
		    $to   =~ /(\d+\.\d+\.\d+\.\d+)/;
		    $info = $1;
		    $from =~ /(\d+\.\d+\.\d+\.\d+)/;
		    $info = "$1 -> $info";
		    $info = translate($info);
		}

		if ($flag =~ /udp/i) {
		    $udp_data[$udp_nr] += $data_sec if ($time >= $ignorefirstseconds);
		    $udp_sum[$prev_time] += $data_sec;
		    if ($time > $udp_lt[$udp_nr]) {
			$udp_lt[$udp_nr] = $time;
		    }
		} else {
		    $tcp_data[$tcp_nr] += $data_sec if ($time >= $ignorefirstseconds);
		    $tcp_sum[$prev_time] += $data_sec;
		    if ($time > $tcp_lt[$tcp_nr]) {
			$tcp_lt[$tcp_nr] = $time;
		    }
		}

		if ($smoothing > 1) {
		    # smoothe by averaging over some time intervals
		    shift @sm_array;
		    push @sm_array, $data_sec;
		    $tmp_smdata = 0;
		    for ($j = 0; $j < $smoothing; ++$j) {
			$tmp_smdata += $sm_array[$j];
		    }
		    printf outfile "%i %i\n", $prev_time, $tmp_smdata / $smoothing / $secondsperinterval;
		    for ($i = $prev_time + $secondsperinterval; $i < $time ; $i += $secondsperinterval) { 
			shift @sm_array;
			push @sm_array, 0;
			$tmp_smdata = 0;
			for ($j = 0; $j < $smoothing; ++$j) {
			    $tmp_smdata += $sm_array[$j];
			}
			printf outfile "%i %i\n", $i, $tmp_smdata / $smoothing / $secondsperinterval;
		    }
		} else {
		    printf outfile "%i %i\n", $prev_time, $data_sec / $secondsperinterval;
		    for ($i = $prev_time + $secondsperinterval; $i < $time ; $i += $secondsperinterval) { 
			printf outfile "%i 0\n", $i;
		    }
		}
		$prev_time = $time;
		$data_sec = 0;
		$data_pkts = 0;
	    }

	    $data_sec += $size;
	    ++$data_pkts;
	} # if port
    }
    close(outfile);
    $max_time = $time if $time > $max_time;
}

#
# summarize
#
if ($udp_nr > 0) {
    $sum_nr = 0;
    $tmp_sum = 0;
    open(outfile, ">udpavg.log.$cur_nr");
    for ($i = 0; $i < @udp_sum; ++$i) {
	if ($i % $secondsperinterval == 0) {
	    printf outfile "%i %i\n", $i, ($sum_nr == 0) ? 0 : $tmp_sum / $sum_nr;
	    $sum_nr = 0;
	    $tmp_sum = 0;
	}
	$tmp_sum += $udp_sum[$i] / $udp_nr;
	++$sum_nr;
    }
    close(outfile);
}
if ($tcp_nr > 0) {
    $sum_nr = 0;
    $tmp_sum = 0;
    open(outfile, ">tcpavg.log.$cur_nr");
    for ($i = 0; $i < @tcp_sum; ++$i) {
	if ($i % $secondsperinterval == 0) {
	    printf outfile "%i %i\n", $i, ($sum_nr == 0) ? 0 : $tmp_sum / $sum_nr;
	    $sum_nr = 0;
	    $tmp_sum = 0;
	}
	$tmp_sum += $tcp_sum[$i] / $tcp_nr;
	++$sum_nr;
    }
    close(outfile);
}

$udpinfo = "";
$tcpinfo = "";
if ($udp_nr > 0) {
    $udp_dmin = 1000000;
    $udp_davg = 0;
    $udp_dmax = 0;
    open(outfile, ">udp_flow.log.$cur_nr");
    for ($i = 1; $i <= $udp_nr; $i++) {
	# check if time interval has positive length
	if ($udp_lt[$i] - $udp_ft[$i] > 0) {
	    $udp_data[$i] /= $udp_lt[$i] - $udp_ft[$i];
	    print outfile "$udp_data[$i]\n";    
	    $udp_dmin = $udp_data[$i] if $udp_dmin > $udp_data[$i];
	    $udp_dmax = $udp_data[$i] if $udp_dmax < $udp_data[$i];
	    $udp_davg += $udp_data[$i];
	} else {
	    $udp_data[$i] = 0;
	}
    }
    close(outfile);
    $udp_davg /= $udp_nr;
    $udpinfo = sprintf("tfrc: %6.3f byte/s [%2.0f, %2.0f], friendliness %3.2f(intra)",
		       $udp_davg, $udp_dmin, $udp_dmax, ($udp_dmax > 0) ? $udp_dmin / $udp_dmax : 0);
}
if ($tcp_nr > 0) {
    $tcp_dmin = 1000000;
    $tcp_davg = 0;
    $tcp_dmax = 0;
    open(outfile, ">tcp_flow.log.$cur_nr");
    for ($i = 1; $i <= $tcp_nr; $i++) {
	if ($tcp_lt[$i] - $tcp_ft[$i] > 0) {
	    #$tcp_data[$i] /= 60;
	    $tcp_data[$i] /= $tcp_lt[$i] - $tcp_ft[$i];
	    print outfile "$tcp_data[$i]\n";    
	    $tcp_dmin = $tcp_data[$i] if $tcp_dmin > $tcp_data[$i];
	    $tcp_dmax = $tcp_data[$i] if $tcp_dmax < $tcp_data[$i];
	    $tcp_davg += $tcp_data[$i];
	} else {
	    $tcp_data[$i] = 0;
	}
    }
    close(outfile);
    $tcp_davg /= $tcp_nr;
    $tcpinfo = sprintf("tcp: %6.3f byte/s [%2.0f, %2.0f], friendliness %3.2f(intra)",
		       $tcp_davg, $tcp_dmin, $tcp_dmax, ($tcp_dmax > 0) ? $tcp_dmin / $tcp_dmax : 0);
}
if (($udp_nr > 0) && ($tcp_nr > 0) && ($udp_davg + $tcp_davg > 0)) {
    $udpinfo .= sprintf(", %3.1f%%(inter)", 100 * $tcp_davg / ($udp_davg + $tcp_davg));
    $tcpinfo .= sprintf(", %3.1f%%(inter)", 100 * $udp_davg / ($udp_davg + $tcp_davg));
}
open(outfile, ">info$cur_nr.txt");
printf outfile "$info, tcp: $tcp_nr tfrc: $udp_nr, duration: $max_time\\n$udpinfo\\n$tcpinfo";
close(outfile);
