#!/usr/bin/perl
$cur_nr = $ARGV[0];
die "1.st argument must be the tracefile number" if ($cur_nr !~ /^\d+$/);
$pref = "0" if ($cur_nr < 100);
$pref = "00" if ($cur_nr < 10);

open(inf, "info$cur_nr.txt");
@info = <inf>;
close(inf);

$info[0] =~ /tcp: (\d+)/;
$tcp_nr = $1;
$info[0] =~ /tfrc: (\d+)/;
$udp_nr = $1;
$info[0] =~ /duration: (\d+)/;
$duration = $1;
#$duration = 180;

$adjust_scale = 0.0; # adjust scale of graph to avg. throughput, use 0.0 or 2.0
$info[0] =~ /ntfrc.*?\[\d+,\s*(\d+)/;
$max_avg = $1;
$info[0] =~ /ntcp.*?\[\d+,\s*(\d+)/;
$max_avg = ($1 > $max_avg) ? $1 : $max_avg;
$max_avg = $max_avg / 1000 * $adjust_scale + 1;

$start_offs = 0;         # affects graph xrange
$end_offs = $duration;

# number of intervals to compute the future loss rate
# (to compare with current loss prediction)
$num_losspredict_intervals = 4;
$ignorefirstseconds = 60; # affects loss error avg. and std. dev.

print "plotting $ARGV[0]...\n";
for ($i = 1; $i < @ARGV; ++$i) {
    $average = 1 if ($ARGV[$i] eq "-avg");
    $p_short = 1   if ($ARGV[$i] eq "-short");
    $single_ps = 1 if ($ARGV[$i] eq "-single");
}

open(cmd, ">cmd.tmp");
if ($p_short) {
    #printf cmd "set term post color solid 20\n";
    #printf cmd "set output \"br$pref$cur_nr.ps\"\n";
    printf cmd "set term post eps color solid 20\n";
    printf cmd "set output \"bitrate$pref$cur_nr.ps\"\n";
} else {
    printf cmd "set term post color solid 20\n";
    printf cmd "set output \"bitrate$pref$cur_nr.tmp\"\n";
}
printf cmd "set title \"@info\"\n";
printf cmd "set xlabel \"time (s)\"\n";
printf cmd "set xrange [$start_offs:$end_offs]\n";
printf cmd "set ylabel \"throughput (KByte/s)\"\n";
if ($adjust_scale > 0) {
    printf cmd "set yrange [0:$max_avg]\n";
} else {
    printf cmd "set yrange [0:]\n";
}
printf cmd "plot ";
if (!$average) {
    for ($i = 1; $i <= $tcp_nr; $i++) {
	printf cmd ", " if ($i > 1);
	if ($i == 1) {
	    printf cmd "\"tcp$i.log.$cur_nr\" u 1:(\$2/1000) t \"TCP\" w l 1";
	} else {
	    printf cmd "\"tcp$i.log.$cur_nr\" u 1:(\$2/1000) notitle w l 1";
	}
    }
    for ($i = 1; $i <= $udp_nr; $i++) {
	printf cmd ", " if ($i > 1 || $tcp_nr > 0);
	if ($i == 1) {
	    printf cmd "\"udp$i.log.$cur_nr\" u 1:(\$2/1000) t \"TFRC\" w l 2";
	} else {
	    printf cmd "\"udp$i.log.$cur_nr\" u 1:(\$2/1000) notitle w l 2";
	}
    }
} else {
    printf cmd "\"tcpavg.log.$cur_nr\" u 1:(\$2/1000) t \"TCP\" w l 1" if ($tcp_nr > 0) ;
    printf cmd ", " if ($tcp_nr > 0 && $udp_nr > 0);
    printf cmd "\"udpavg.log.$cur_nr\" u 1:(\$2/1000) t \"TFRC\" w l 2" if ($udp_nr > 0);
}
printf cmd "\n";
close(cmd);
system "gnuplot cmd.tmp";

if ($p_short) {
    system "rm -f *.tmp";
    exit;
}

#if ($udp_nr > 0 && $tcp_nr > 0) {
#    open(cmd, ">cmd1.tmp");
#    printf cmd "set term post color solid 20\n";
#    printf cmd "set output \"diff$pref$cur_nr.ps\"\n";
#    printf cmd "set title \"tcp - tfm throughput\"\n";
#    printf cmd "set xzeroaxis\n";
#    printf cmd "set xlabel \"time (s)\"\n";
#    printf cmd "set xrange [$start_offs:$end_offs]\n";
#    printf cmd "set ylabel \"throughput diff.(KByte/s)\"\n";
#    printf cmd "plot \"diff.log.$cur_nr\" u 1:(\$2/1000) t \"tcpavg - udpavg\" w l\n";
#    close(cmd);
#    system "gnuplot cmd1.tmp";
#}

for ($i = 1; $i <= $udp_nr; $i++) {
    open(cmd, ">cmd3.tmp");
    printf cmd "set term post color solid 20\n";
    printf cmd "set output \"bexp$i.tmp\"\n";
    printf cmd "set xlabel \"time (s)\"\n";
    printf cmd "set xrange [$start_offs:$end_offs]\n";
    printf cmd "set ylabel \"bitrate (KByte/s)\"\n";
    printf cmd "plot \"all$i.log.$cur_nr\" u 1:(\$5/1000) t \"actual bitrate $i.$cur_nr\" w l";
    printf cmd ", \"all$i.log.$cur_nr\" u 1:(\$6/1000) t \"expected bitrate $i.$cur_nr\" w l";
    # printf cmd ", \"all$i.log.$cur_nr\" u 1:(\$7/1000) t \"reported bitrate $i.$cur_nr\" w l";
    printf cmd "\n";
    close(cmd);
    system "gnuplot cmd3.tmp";
}

for ($i = 1; $i <= $udp_nr; $i++) {
    open(cmd, ">cmd4.tmp");
    printf cmd "set term post color solid 20\n";
    printf cmd "set output \"rtt$i.tmp\"\n";
    printf cmd "set title \"($i.$cur_nr) rtt, tzero\"\n";
    printf cmd "set xlabel \"time (s)\"\n";
    printf cmd "set xrange [$start_offs:$end_offs]\n";
    printf cmd "set yrange [0:]\n";
    printf cmd "set ylabel \"rtt (ms)\"\n";
    printf cmd "plot ";
    printf cmd "\"all$i.log.$cur_nr\" u 1:9 s u t \"rtt\" w l 1, ";
    printf cmd "\"all$i.log.$cur_nr\" u 1:8 s u t \"rtt (ema)\" w l 2, ";
    printf cmd "\"all$i.log.$cur_nr\" u 1:10 s u t \"tzero\" w l 3\n";
    close(cmd);
    system "gnuplot cmd4.tmp";
}

@error_avg = ();
for ($ix = 1; $ix <= $udp_nr; $ix++) {
    @loss_seq   = ();
    @loss_time  = ();
    @loss_size  = ();
    @loss_start = ();
    @loss_end   = ();

    open(infile, "losshist$ix.log.$cur_nr") or die "can't open losshist$ix.log.$cur_nr\n";
    while ($line = <infile>) {
	chomp $line;
	($time, $num_loss, $seq_no) = split(/\s+/, $line);
	push @loss_seq,   $seq_no;
	push @loss_time,  $time;
	push @loss_size,  $num_loss;
	push @loss_start, -1;
	push @loss_end,   -1;
	
    }
    close(infile);

    $max_seq = 0;
    $min_seq = $seq_no;

    open(infile, "all$ix.log.$cur_nr") or die "can't open all$ix.log.$cur_nr\n";
    open(outfile, ">pred_err$ix.log.$cur_nr");
    $count = 0;
    $error_avg[$ix] = 0;

    $cur_loss_iv = 0;

    $start_idx = 0; # to speed up assignment of loss windows to loss events

    while ($line = <infile>) {
	chomp $line;
	($time, $seq_no, $round_no, $l_act, $b_act, $b_exp, $b_rep, $rtt_ema, $rtt_cur, $tzero, $ppl_start, $ppl_end) = split(/\s+/, $line);

	$min_seq = $ppl_start if ($min_seq > $ppl_start);	
	$max_seq = $ppl_end   if ($max_seq < $ppl_end);

	for ($i = $start_idx; $i < @loss_seq; ++$i) {
	    if ($ppl_start > $loss_seq[$i] && $loss_end[$i] == -1) {
		$loss_end[$i] = $time;
		$start_idx = $i; # older losses are already set
	    }
	    if ($ppl_end  >= $loss_seq[$i] && $loss_start[$i] == -1) {
		$loss_start[$i] = $time;
	    }
	}

        # diff. between loss estimate and real loss (packets per loss event)
	# last loss interval ignored (no loss at end)
	if ($time >= $loss_time[$cur_loss_iv + 1]) {
	    ++$cur_loss_iv; # report belongs to next loss interval
	}

	if (($l_act > 0.0) && ($cur_loss_iv < @loss_size - 1)) {
	    $tmp_num = 0;
	    $tmp_ival = 0;

	    for ($i = $cur_loss_iv; ($i < $cur_loss_iv + $num_losspredict_intervals)
		 && ($i < @loss_size); ++$i) {
		#print $i + $cur_loss_iv, " ";
		++$tmp_num;
		$tmp_ival += $loss_size[$i];
	    }
	    #print "\n";
	    # compute average ppl for the next for loss intervals
	    if ($tmp_num > 0) {
		$tmp_ival /= $tmp_num;
		printf outfile "%f %f %f\n", $time, abs($l_act - 1 / $tmp_ival), 1 / $tmp_ival;
		if ($time > $ignorefirstseconds) {
		    $error_avg[$ix] += abs($l_act - 1 / $tmp_ival);
		    ++$count;
		}
	    }
	}
    }

    for ($i = 0; $i < @loss_seq; ++$i) {
	$loss_end[$i] = $time if ($loss_end[$i] == -1);
    }

    if ($count > 0) {
	$error_avg[$ix] /= $count;
    } else {
	print STDERR "ERROR: couldn't comput loss average, std. dev. (no losses in time interval)\n";
	$error_avg[$ix] = 0;
    }
    close(outfile);
    close(infile);

    open(outfile, ">losswin$ix.log.$cur_nr");
    for ($i = 0; $i < @loss_seq; ++$i) {
	print outfile "$loss_start[$i] $loss_seq[$i]\n$loss_end[$i] $loss_seq[$i]\n\n";
    }
    close(outfile);
    
    open(cmd, ">cmd5.tmp");
    print cmd "set term post color solid 20\n";
    print cmd "set output \"win$ix.tmp\"\n";
    print cmd "set title \"($ix.$cur_nr) loss window\"\n";
    print cmd "set xlabel \"time (s)\"\n";
    print cmd "set xrange [$start_offs:$end_offs]\n";
    print cmd "set xzeroaxis\n";
    print cmd "set ylabel \"sequence number\"\n";
    print cmd "set multiplot\n";
    print cmd "set yrange [$min_seq:", $max_seq + 500, "]\n";
    print cmd "plot \"all$ix.log.$cur_nr\" u 1:11 notitle w l 1\n";
    print cmd "plot \"all$ix.log.$cur_nr\" u 1:12 notitle w l 1\n";
    print cmd "plot \"losswin$ix.log.$cur_nr\" t \"loss events $ix.$cur_nr\" w l 3\n";
    close(cmd);
    system "gnuplot cmd5.tmp";
}

open(outfile, ">pred_err.log.$cur_nr");
for ($ix = 1; $ix <= $udp_nr; $ix++) {
    $error_sd = 0;
    $count = 0;
    open(infile, "pred_err$ix.log.$cur_nr");
    while ($line = <infile>) {
	chomp $line;
	($time, $est_error) = split(/\s+/, $line);
	if ($time > $ignorefirstseconds) {
	    $error_sd += ($est_error - $error_avg[$ix]) * ($est_error - $error_avg[$ix]);
	    ++$count;
	}
    }

    if ($count > 1) {
	$error_sd = sqrt($error_sd / ($count - 1));
    } else {
	if ($count == 0) {
	    $error_avg[$ix] = 0;
	}
	$error_sd = 0;
    }
    close(infile);
    
    print outfile "$ix $error_avg[$ix] $error_sd\n";

    open(cmd, ">cmd2.tmp");
    printf cmd "set term post color solid 20\n";
    printf cmd "set output \"loss$ix.tmp\"\n";
    printf cmd "set title \"($ix.$cur_nr) lossrate\\n";
    printf cmd "error: avg. %7.6f, std. dev. %7.6f\"\n", $error_avg[$ix], $error_sd;
    printf cmd "set xlabel \"time (s)\"\n";
    printf cmd "set xrange [$start_offs:$end_offs]\n";
    printf cmd "set ylabel \"lossrate\"\n";
    printf cmd "plot \"pred_err$ix.log.$cur_nr\" u 1:3 t \"actual loss rate\" w steps 1, ";
    printf cmd "     \"all$ix.log.$cur_nr\" u 1:4 s u t \"est. loss rate\" w l 2, ";
    #printf cmd ", \"losshist$ix.log.$cur_nr\" u 1:(\$3==3 ? 0 : 1/0) t \"loss events\" w p\n";
    printf cmd "     \"losshist$ix.log.$cur_nr\" u 1:(0) t \"loss events\" w p 3\n";
    close(cmd);
    system "gnuplot cmd2.tmp";
}
close(outfile);

if ($single_ps) {
    for ($ix = 1; $ix <= $udp_nr; $ix++) {
	system "mv bexp$ix.tmp bexp$ix" . "_$pref$cur_nr.ps";
	system "mv loss$ix.tmp loss$ix" . "_$pref$cur_nr.ps";
	system "mv win$ix.tmp win$ix" . "_$pref$cur_nr.ps";
	system "mv rtt$ix.tmp rtt$ix" . "_$pref$cur_nr.ps";
	system "mv bitrate$pref$cur_nr.tmp bitrate$pref$cur_nr.ps";
    }
} else {
    $mergecmd = "mergeps.pl graph.tmp ";
    for ($ix = 1; $ix <= $udp_nr; $ix++) {
	$mergecmd = "$mergecmd bexp$ix.tmp loss$ix.tmp win$ix.tmp rtt$ix.tmp";
    }
    $mergecmd = "$mergecmd";
    system $mergecmd;
    system "pstops '2:0R@.69(0cm,27cm)+1R@.69(0cm,14.5cm)' graph.tmp graph2.tmp";
    system "mergeps.pl graph$pref$cur_nr.ps bitrate$pref$cur_nr.tmp graph2.tmp";
    system "mv bitrate$pref$cur_nr.tmp br$pref$cur_nr.ps";
}
system "rm -f *.tmp";
