#!/usr/bin/perl
$cur_nr = shift;
die "1.st argument must be the tracefile number" if ($cur_nr !~ /^\d+$/);
$pref = "0" if ($cur_nr < 100);
$pref = "00" if ($cur_nr < 10);

$start_offs = shift;         # affects graph xrange
$end_offs = shift;

open(inf, "info$cur_nr.txt");
@info = <inf>;
close(inf);

$info[0] =~ /tcp: (\d+)/;
$tcp_nr = $1;
$info[0] =~ /tfrc: (\d+)/;
$udp_nr = $1;
$info[0] =~ /duration: (\d+)/;
$duration = $1;
#$duration = 180;

$end_offs = $duration if ($end_offs == 0);

# number of intervals to compute the future loss rate
# (to compare with current loss prediction)
$num_losspredict_intervals = 4;
$ignorefirstseconds = 60; # affects loss error avg. and std. dev.
$average = 0;

#$combined = 1; # combined RTT, loss rate graphs

for ($i = 0; $i < @ARGV; ++$i) {
    $m_bitrate = 1   if ($ARGV[$i] eq "bitrate");
    $m_bexp = 1 if ($ARGV[$i] eq "bexp");
    $m_loss = 1 if ($ARGV[$i] eq "loss");
    $m_losswin = 1 if ($ARGV[$i] eq "losswin");
    $m_rtt = 1 if ($ARGV[$i] eq "rtt");
    $m_diff = 1 if ($ARGV[$i] eq "diff"); # change colors
    $m_diff2 = 1 if ($ARGV[$i] eq "diff2"); # change colors
}

if ($m_bitrate) {
    open(cmd, ">cmd.tmp");
    printf cmd "set term post eps color solid 20\n";
    printf cmd "set output \"bitrate$pref$cur_nr.ps\"\n";
    # printf cmd "set title \"@info\"\n";
    printf cmd "set xlabel \"time (s)\"\n";
    printf cmd "set xrange [$start_offs:$end_offs]\n";
    printf cmd "set ylabel \"throughput (KByte/s)\"\n";
    printf cmd "set yrange [0:]\n";
    printf cmd "set offsets 0, 0, 10, 0\n";
    printf cmd "plot ";
    if (not $average) {
	for ($i = 1; $i <= $tcp_nr; $i++) {
	    printf cmd ", " if ($i > 1);
	    if ($m_diff == 1) {
		print cmd "\"tcp$i.log.$cur_nr\" u 1:(\$2/1000) t \"TCP $i\" w l ", $i;
	    } elsif ($m_diff2 == 1) {
		print cmd "\"tcp$i.log.$cur_nr\" u 1:(\$2/1000) t \"TCP $i\" w l ", $i*2 - 1;
	    } else {
		if ($i == 1) {
		    printf cmd "\"tcp$i.log.$cur_nr\" u 1:(\$2/1000) t \"TCP\" w l 1";
		} else {
		    printf cmd "\"tcp$i.log.$cur_nr\" u 1:(\$2/1000) notitle w l 1";
		}
	    }
	}
	for ($i = 1; $i <= $udp_nr; $i++) {
	    printf cmd ", " if ($i > 1 || $tcp_nr > 0);
	    if ($m_diff == 1) {
		print cmd "\"udp$i.log.$cur_nr\" u 1:(\$2/1000) t \"TFRC $i\" w l ", $i + $tcp_nr;
	    } elsif ($m_diff2 == 1) {
		print cmd "\"udp$i.log.$cur_nr\" u 1:(\$2/1000) t \"TFRC $i\" w l ", $i*2;
	    } else {
		if ($i == 1) {
		    print cmd "\"udp$i.log.$cur_nr\" u 1:(\$2/1000) t \"TFRC\" w l 2";
		} else {
		    print cmd "\"udp$i.log.$cur_nr\" u 1:(\$2/1000) notitle w l 2";
		}
	    }
	}
    } else {
	printf cmd "\"tcpavg.log.$cur_nr\" u 1:(\$2/1000) t \"TCP\" w l 1" if ($tcp_nr > 0) ;
	printf cmd ", " if ($tcp_nr > 0 && $udp_nr > 0);
	printf cmd "\"udpavg.log.$cur_nr\" u 1:(\$2/1000) t \"TFRC\" w l 2" if ($udp_nr > 0);
    }
    printf cmd "\n";
    close(cmd);
    system "gnuplot cmd.tmp";
}

if ($m_bexp) {
    for ($i = 1; $i <= $udp_nr; $i++) {
	open(cmd, ">cmd3.tmp");
	printf cmd "set term post eps color solid 20\n";
	printf cmd "set output \"bexp$i" . "_$pref$cur_nr.ps\"\n";
	printf cmd "set xlabel \"time (s)\"\n";
	printf cmd "set xrange [$start_offs:$end_offs]\n";
	printf cmd "set ylabel \"bitrate (KByte/s)\"\n";
	printf cmd "plot \"all$i.log.$cur_nr\" u 1:(\$5/1000) t \"actual bitrate $i.$cur_nr\" w l";
	printf cmd ", \"all$i.log.$cur_nr\" u 1:(\$6/1000) t \"expected bitrate $i.$cur_nr\" w l";
	# printf cmd ", \"all$i.log.$cur_nr\" u 1:(\$7/1000) t \"reported bitrate $i.$cur_nr\" w l";
	printf cmd "\n";
	close(cmd);
	system "gnuplot cmd3.tmp";
    }
}

if ($m_rtt) {
    for ($i = 1; $i <= $udp_nr; $i++) {
	open(cmd, ">cmd4.tmp");
	printf cmd "set term post epscolor solid 20\n";
	printf cmd "set output \"rtt$i" . "_$pref$cur_nr.ps\"\n";
	printf cmd "set title \"($i.$cur_nr) rtt, tzero\"\n";
	printf cmd "set xlabel \"time (s)\"\n";
	printf cmd "set xrange [$start_offs:$end_offs]\n";
	printf cmd "set yrange [0:]\n";
	printf cmd "set ylabel \"rtt (ms)\"\n";
	printf cmd "plot ";
	printf cmd "\"all$i.log.$cur_nr\" u 1:8 s u t \"rtt (ema)\", ";
	printf cmd "\"all$i.log.$cur_nr\" u 1:9 s u t \"rtt\", ";
	printf cmd "\"all$i.log.$cur_nr\" u 1:10 s u t \"tzero\" \n";
	close(cmd);
	system "gnuplot cmd4.tmp";
    }
}

if ($m_losswin) {
    @error_avg = ();
    for ($ix = 1; $ix <= $udp_nr; $ix++) {
	@loss_seq   = ();
	@loss_time  = ();
	@loss_size  = ();
	@loss_start = ();
	@loss_end   = ();

	open(infile, "losshist$ix.log.$cur_nr") or die "can't open losshist$ix.log.$cur_nr\n";
	while ($line = <infile>) {
	    chomp $line;
	    ($time, $num_loss, $seq_no) = split(/\s+/, $line);
	    push @loss_seq,   $seq_no;
	    push @loss_time,  $time;
	    push @loss_size,  $num_loss;
	    push @loss_start, -1;
	    push @loss_end,   -1;
	    
	}
	close(infile);

	$max_seq = 0;
	$min_seq = $seq_no;

	open(infile, "all$ix.log.$cur_nr") or die "can't open all$ix.log.$cur_nr\n";
	open(outfile, ">pred_err$ix.log.$cur_nr");
	$count = 0;
	$error_avg[$ix] = 0;

	$cur_loss_iv = 0;

	$start_idx = 0; # speed up assignment of loss windows to loss events

	while ($line = <infile>) {
	    chomp $line;
	    ($time, $seq_no, $round_no, $l_act, $b_act, $b_exp, $b_rep, $rtt_ema, $rtt_cur, $tzero, $ppl_start, $ppl_end) = split(/\s+/, $line);

	    $min_seq = $ppl_start if ($min_seq > $ppl_start);	
	    $max_seq = $ppl_end   if ($max_seq < $ppl_end);

	    for ($i = $start_idx; $i < @loss_seq; ++$i) {
		if ($ppl_start > $loss_seq[$i] && $loss_end[$i] == -1) {
		    $loss_end[$i] = $time;
		    $start_idx = $i; # older losses are already set
		}
		if ($ppl_end  >= $loss_seq[$i] && $loss_start[$i] == -1) {
		    $loss_start[$i] = $time;
		}
	    }

	    # diff. between loss estimate and real loss (packets per loss event)
	    # last loss interval ignored (no loss at end)
	    if ($time >= $loss_time[$cur_loss_iv + 1]) {
		++$cur_loss_iv; # report belongs to next loss interval
	    }

	    if (($l_act > 0.0) && ($cur_loss_iv < @loss_size)) {
		$tmp_num = 0;
		$tmp_ival = 0;

		for ($i = $cur_loss_iv; ($i < $cur_loss_iv + $num_losspredict_intervals)
		     && ($i < @loss_size); ++$i) {
		    #print $i + $cur_loss_iv, " ";
		    ++$tmp_num;
		    $tmp_ival += $loss_size[$i];
		}
		#print "\n";
		# compute average ppl for the next for loss intervals
		if ($tmp_num > 0) {
		    $tmp_ival /= $tmp_num;
		    printf outfile "%f %f %f\n", $time, abs($l_act - 1 / $tmp_ival), 1 / $tmp_ival;
		    if ($time > $ignorefirstseconds) {
			$error_avg[$ix] += abs($l_act - 1 / $tmp_ival);
			++$count;
		    }
		}
	    }
	}

	for ($i = 0; $i < @loss_seq; ++$i) {
	    $loss_end[$i] = $time if ($loss_end[$i] == -1);
	}

	if ($count > 0) {
	    $error_avg[$ix] /= $count;
	} else {
	    print STDERR "ERROR: couldn't comput loss average, std. dev. (no losses in time interval)\n";
	    $error_avg[$ix] = 0;
	}
	close(outfile);
	close(infile);

	open(outfile, ">losswin$ix.log.$cur_nr");
	for ($i = 0; $i < @loss_seq; ++$i) {
	    print outfile "$loss_start[$i] $loss_seq[$i]\n$loss_end[$i] $loss_seq[$i]\n\n";
	}
	close(outfile);
	
	open(cmd, ">cmd5.tmp");
	print cmd "set term post eps color solid 20\n";
	print cmd "set output \"win$ix" . "_$pref$cur_nr.ps\"\n";
	print cmd "set title \"($ix.$cur_nr) loss window\"\n";
	print cmd "set xlabel \"time (s)\"\n";
	print cmd "set xrange [$start_offs:$end_offs]\n";
	print cmd "set xzeroaxis\n";
	print cmd "set ylabel \"sequence number\"\n";
	print cmd "set multiplot\n";
	print cmd "set yrange [$min_seq:", $max_seq + 500, "]\n";
	print cmd "plot \"all$ix.log.$cur_nr\" u 1:11 notitle w l 1\n";
	print cmd "plot \"all$ix.log.$cur_nr\" u 1:12 notitle w l 1\n";
	print cmd "plot \"losswin$ix.log.$cur_nr\" t \"loss events $ix.$cur_nr\" w l 3\n";
	close(cmd);
	system "gnuplot cmd5.tmp";
    }
}

if ($m_loss) {
    open(outfile, ">pred_err.log.$cur_nr");
    for ($ix = 1; $ix <= $udp_nr; $ix++) {
	$error_sd = 0;
	$count = 0;
	open(infile, "pred_err$ix.log.$cur_nr");
	while ($line = <infile>) {
	    chomp $line;
	    ($time, $est_error) = split(/\s+/, $line);
	    if ($time > $ignorefirstseconds) {
		$error_sd += ($est_error - $error_avg[$ix]) * ($est_error - $error_avg[$ix]);
		++$count;
	    }
	}

	if ($count > 1) {
	    $error_sd = sqrt($error_sd / ($count - 1));
	} else {
	    if ($count == 0) {
		$error_avg[$ix] = 0;
	    }
	    $error_sd = 0;
	}
	close(infile);
	
	print outfile "$ix $error_avg[$ix] $error_sd\n";

	open(cmd, ">cmd2.tmp");
	printf cmd "set term post eps color solid 20\n";
	printf cmd "set output \"loss$ix" . "_$pref$cur_nr.ps\"\n";
	#printf cmd "set title \"($ix.$cur_nr) lossrate\\n";
	#printf cmd "error: avg. %7.6f, std. dev. %7.6f\"\n", $error_avg[$ix], $error_sd;
	printf cmd "set lmargin 10\n";
	printf cmd "set rmargin 8\n";
	printf cmd "set xlabel \"time (s)\"\n";
	printf cmd "set xrange [$start_offs:$end_offs]\n";
	printf cmd "set ylabel \"lossrate\"\n";
	printf cmd "plot \"all$ix.log.$cur_nr\" u 1:4 s u t \"est. loss rate\" w l";
	printf cmd ", \"pred_err$ix.log.$cur_nr\" u 1:3 t \"actual loss rate\" w steps";
	#printf cmd ", \"losshist$ix.log.$cur_nr\" u 1:(\$3==3 ? 0 : 1/0) t \"loss events\" w p\n";
	printf cmd ", \"losshist$ix.log.$cur_nr\" u 1:(0) t \"loss events\" w p 3\n";
	close(cmd);
	system "gnuplot cmd2.tmp";
    }
    close(outfile);
}
system "rm -f *.tmp";
