#!/usr/bin/perl

sub calc_time {
    local ($fhh, $fmm, $fss, $hh, $mm, $ss) = @_ ; 
    local ($diff) ; 

    if ($hh < $fhh) { # a day change occured
	$diff = 24 ; 
    }
    else {
	$diff = 0 ; 
    }

    $diff = $diff + ($hh - $fhh) ;
    $diff = $diff*60 ; 
    $diff = $diff + ($mm - $fmm) ; 
    $diff = $diff*60 ; 
    $diff = $diff + ($ss - $fss); 

    return $diff ; 
}

sub translate {
    local ($line) = @_;

    $line =~ s/128\.119\.41\.224/umass/g;
    $line =~ s/128\.119\.241\.136/umass/g;
    $line =~ s/134\.155\.48\.96/unima/g;
    $line =~ s/128\.16\.64\.73/ucl/g;
    $line =~ s/208\.219\.91\.202/nokia/g;
    $line =~ s/24\.6\.255\.61/cable-modem/g;
    $line =~ s/192\.150\.187\.31/aciri\(lemming\)/g ;
    $line =~ s/192\.150\.187\.61/aciri\(hyrax\)/g ;
    $line =~ s/192\.150\.187\.23/aciri\(ox\)/g ;
    $line =~ s/192\.150\.187\.20/aciri\(aak\)/g ;
    $line =~ s/192\.150\.187\.12/aciri\(griffin\)/g ;
    return $line;
}

if (@ARGV < 1) {
    printf "tcp-loss.pl cur_nr\n";
    exit;
}

$firstline = 1;

$cur_nr = shift;

open(inf, "info$cur_nr.txt");
@info = <inf>;
close(inf);

$info[0] =~ /tcp: (\d+)/;
$tcp_nr = $1;
$info[0] =~ /tfrc: (\d+)/;
$udp_nr = $1;

exit if ($tcp_nr == 0);

$cum_t = 0;
$nr_t = 0;

@ports = ();
for ($i = 0; $i < $tcp_nr; ++$i) {
    push @ports, 20058 + $i * 100;
}

if (not open(fp, "net.log.$cur_nr")) {
    if (not open(fp, "net.log.$cur_nr.gz")) {
	die "couldn't open net.log.$cur_nr.gz file";
    } else {
	close(fp);
	open(fp, "gzip -dc < net.log.$cur_nr.gz |");
    }
}
@log = <fp>;
close(fp);

if (open(rttfile, "rtt_avg.log.$cur_nr")) {
    $rtt_avg = <rttfile>;
    $rtt_nr = 1;
    close(rttfile);
} else {
    $id = 1;
    while (1) {
	if (not open(infile, "all$id.log.$cur_nr")) {
	    if (not open(infile, "all$id.log.$cur_nr.gz")) {
		last;
	    } else {
		close(infile);
		open(infile, "gzip -dc < all$id.log.$cur_nr.gz |");
	    }
	}
	
	while ($line = <infile>) {
	    chomp $line;
	    @entry = split(/\s+/, $line);
	    ++$rtt_nr;
	    $rtt_avg += $entry[8];
	}
	++$id;
	close(infile);
    }
    open(rttfile, ">rtt_avg.log.$cur_nr");
    $rtt_avg /= $rtt_nr;
    print rttfile $rtt_avg, "\n";
    close(rttfile);
}

$dup = 0;
$tout = 0;
$td = 0;
$tot_nr = 0;

open(outfile, ">tcp_loss.log.$cur_nr") or die "can't open file tcp_loss.log.$cur_nr\n";
for ($port = 0; $port < @ports; $port++) {
    #printf "port %s\n", $ports[$port];
    $prev_time = 0;
    $high_ack = 0;
    $maxseq = -1;
    $lastend = -1;
    $laststart = -1;
    $lasttime = -1;
    $gaptime = -1;
    $tcp_packets = 0;
    $tdack = 0;
    $timeout = 0;
    %seq_time = ();
    %seq_no = ();
    %ack_no = ();

    foreach $line (@log) {
	($hh,$mm,$ss,$from,$to,$flag) = split(/[\s>:\(\)]+/,$line); 
	if ($flag =~ /\.|P/) {
	    # reparse TCP
	    ($hh, $mm, $ss, $from, $to, $flag, $start, $end, $size) = 
		split (/[\s>:\(\)]+/, $line);
	} else {
	    next;
	}
	#
	# need to check for firstline to convert absolute time 
	# into relative
	#
	if ($firstline == 1) {
	    $fhh = $hh ;
	    $fmm = $mm ; 
	    $fss = $ss ; 
	    $firstline = 0;
	}

	$time = calc_time($fhh, $fmm, $fss, $hh, $mm, $ss);
	
	# check for ACK
	if (($from =~ /$ports[$port]/) && ($start =~ /ack/i)) {
	    # receiver ACK
	    if ($ack_no{$end} == undef) {
		$ack_no{$end} = 1;
	    } else {
		++$ack_no{$end};
		if ($ack_no{$end} == 4) { # account only for 3rd dup ACK !
		    ++$tdack;
		    ++$td;
		    #printf "%6.2f: ACK $end ($ack_no{$end})\n", $time;
		}
	    }
	    $high_ack = $end;
	    next;
	}

	if ($to =~ /$ports[$port]/) {
	    if ($start =~ /ack/i) {
		next; #sender ACK
	    }

	    # check seq_no
	    if ($seq_no{$start} == undef) {
		if ($start >= $maxseq) {
		    # first seen, ok
		} else {
		    # was missing
		    die "$time: gap error" if ($gap_time == -1);
		    #printf "%6.2f: missing packet ($end) arrived after %1.4f seconds\n", $time, $time - $gaptime;
		}
		$seq_no{$start} = 1;
		$seq_time{$start} = $time;
	    } else {
		# duplicate
		++$dup;
		++$seq_no{$start};
		#printf "%6.2f: duplicate packet ($start) - %1.4f\n", $time, $time - $seq_time{$start};
	    }

	    if (($time - $lasttime >= 0.25) && ($time - $lasttime >= 4 * $rtt_avg / 1000)) {
		++$timeout;
		++$tout;
		#print $time, ": timeout (", $time - $lasttime, ")\n";
	    }

	    if (($lastend > -1) && ($start != $lastend)) {
		# seq no gap
		#print "$time: gap, expected $lastend, got $start\n";
		$gaptime = $lasttime;
	    }
	    ++$tcp_packets;
	    ++$tot_nr;
	    $maxseq = $end if ($maxseq < $end);
	    $lastend = $end;
	    $laststart = $start;
	    if ($time - $lasttime > 0.3) {
		$cum_t += $time - $lasttime;
		++$nr_t;
		#print $time, ": \t", $time - $lasttime, ", sum: $cum_t, avg: ", $cum_t / $nr_t, "\n";
	    }
	    $lasttime = $time;
	} # if port
    }
    if ($tcp_packets > 9) { # only representative if 10 or more packets
	print outfile $tdack / $tcp_packets, " ", $timeout / $tcp_packets, "\n";
    } else {
	print outfile "-1 -1\n";
    }
}
#print "packets: $tot_nr \t duplicates: $dup, timeouts: $tout, TDACKs: $td\n";

close(outfile);
