#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "config_unix.h"
#include "config_win32.h"
#include "util.h"
#include "lossfraction.h"

double lossfraction_v3() { /* split last loss interval to deweight */
        int i;
        double pplavg, weightsum;
        double ppl_;
        int size = 0;

        /* compute avg. loss interval size without current loss interval */
        pplavg = 0.0; weightsum = 0.0;
        for (i = 0; i < weights; ++i) {
                ppl_ = ppl[(ppl_cur - i + weights) % (weights + 1)];
                if (ppl_ > 0) {
                        weightsum += weight[i];
                        pplavg += ppl_ * weight[i];
                        size += ppl_;
                }
        }
        assert(weightsum > 0);
        assert(pplavg > 0);
        pplavg /= weightsum;

        if (ppl[ppl_cur] <= pplavg) {
                ppl_end = high_offs - ppl[ppl_cur];
                ppl_start = max(ppl_end - size + 1, 0);
                return (1.0 / pplavg);
        }

        /* add fake loss interval when ppl[ppl_cur] > 2 * pplavg */
        if (ppl[ppl_cur] > DW_FACTOR * pplavg) {
                ppl_ = ppl[ppl_cur];
                ppl_cur = (ppl_cur + 1) % (weights + 1);
                ppl[ppl_cur] = ppl_;
        }

        pplavg = 0.0; weightsum = 0.0; size = 0;
        /* compute lossrate with current loss interval */
        for (i = 0; i < weights; ++i) {
                ppl_ = ppl[(ppl_cur - i + weights + 1) % (weights + 1)];
                if (ppl_ > 0) {
                        weightsum += weight[i];
                        pplavg += ppl_ * weight[i];
                        size += ppl_;
                }
        }
        assert(weightsum > 0);
        assert(pplavg > 0);
        pplavg /= weightsum;

        ppl_end = high_offs;
        ppl_start = max(ppl_end - size + 1, 0);
        return (1.0 / pplavg);
}


double lossfraction_v2() { /* use mult to deweight old loss intervals */
        double pplavg1 = 0.0, pplavg2 = 0.0;
        double weightsum1 = 0.0, weightsum2 = 0.0;
        int size1 = 0, size2 = 0, ppl1, ppl2;
        int num_intervals = 0;
        int i;

        /* loss avg. without latest interval */
        for (i = 0; i < weights; ++i) {
                ppl2 = ppl[(ppl_cur - i + weights) % (weights + 1)];
                if (ppl2 > 0) {
                        /* pplavg2 += ppl2 * weight[i] * mult[i]; */
                        /* weightsum2 += weight[i] * mult[i]; */
                        pplavg2 += ppl2 * weight[i];
                        weightsum2 += weight[i];
                        ++num_intervals;
                }
        }

        /* determine deweighting factor */
        if (num_intervals > 0 && weightsum2 > 0 && pplavg2 > 0) {
                pplavg2 /= weightsum2;
                new_mult = min((DW_FACTOR * pplavg2) / ppl[ppl_cur], 1.0);
        } else {
                new_mult = 1.0;
        }

        /* reset for second computation */
        pplavg2 = 0.0;
        weightsum2 = 0.0;

        /* compute lossrate with/without packets after last loss
         * --> only take these packets into account if that decreases the loss rate */
        for (i = 0; i < weights; ++i) {
                ppl1 = ppl[(ppl_cur - i + weights + 1) % (weights + 1)];
                ppl2 = ppl[(ppl_cur - i + weights) % (weights + 1)];
                /* no deweighting of current interval */
                if (ppl1 > 0) {
                        if (i > 0) {
                                /* double m1 = max(mult[i] * new_mult, MIN_MULT); */
                                double m1 = max(new_mult, MIN_MULT);
                                weightsum1 += weight[i] * m1;
                                pplavg1 += ppl1 * weight[i] * m1;
                        } else {
                                weightsum1 += weight[i];
                                pplavg1 += ppl1 * weight[i];
                        }
                        size1 += ppl1;
                }
                if (ppl2 > 0) {
                        if (i > 0) {
                                /* double m2 = max(mult[i + 1] * new_mult, MIN_MULT); */
                                double m2 = max(new_mult, MIN_MULT);
                                weightsum2 += weight[i] * m2;
                                pplavg2 += ppl2 * weight[i] * m2;
                        } else {
                                weightsum2 += weight[i];
                                pplavg2 += ppl2 * weight[i];
                        }
                        size2 += ppl2;
                }
                /* printf("%f \t %f, %f %f \n", pplavg1, pplavg2, weightsum1, weightsum2); */
                /* printf("%f ", max(mult[i] * new_mult, MIN_MULT)); */
        }
        /* printf("\n"); */

        assert(weightsum1 > 0);
        assert(weightsum2 > 0);
        assert(pplavg1 > 0);
        assert(pplavg2 > 0);

        pplavg1 /= weightsum1;
        pplavg2 /= weightsum2;

        /* printf("%f \t %f\n\n", pplavg1, pplavg2); */

        if (pplavg1 > pplavg2) {
                ppl_end = high_offs;
                ppl_start = ppl_end - size1 + 1;
        } else {
                ppl_end = high_offs - ppl[ppl_cur];
                ppl_start = ppl_end - size2 + 1;
        }

        return min(1.0 / pplavg1, 1.0 / pplavg2);
}


double lossfraction_v1() { /* no deweighting */
        int i;
        double pplavg1 = 0.0, pplavg2 = 0.0;
        double weightsum1 = 0.0, weightsum2 = 0.0;
        int size1 = 0, size2 = 0, ppl1, ppl2;

        for (i = 0; i < weights; ++i) {
                /* 
                 * compute lossrate with/without packets after last loss
                 * --> only take these packets into account if that decreases the loss rate
                 */

                ppl1 = ppl[(ppl_cur - i + weights + 1) % (weights + 1)];
                ppl2 = ppl[(ppl_cur - i + weights) % (weights + 1)];
                if (ppl1 > 0) {
                        weightsum1 += weight[i];
                        pplavg1 += ppl1 * weight[i];
                        size1 += ppl1;
                }
                if (ppl2 > 0) {
                        weightsum2 += weight[i];
                        pplavg2 += ppl2 * weight[i];
                        size2 += ppl2;
                }
        }

        assert(weightsum1 > 0);
        assert(weightsum2 > 0);
        assert(pplavg1 > 0);
        assert(pplavg2 > 0);

        pplavg1 /= weightsum1;
        pplavg2 /= weightsum2;

        if (pplavg1 > pplavg2) {
                ppl_end = high_offs;
                ppl_start = ppl_end - size1 + 1;
        } else {
                ppl_end = high_offs - ppl[ppl_cur];
                ppl_start = ppl_end - size2 + 1;
        }

        return min(1.0 / pplavg1, 1.0 / pplavg2);
}

void fake_history(int ppl_est) {
        if (ppl[0] < ppl_est)
                ppl[0] = ppl_est;
}
