#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/param.h>
#include <assert.h>

#include "config_unix.h"
#include "config_win32.h"
#include "net_udp.h"
#include "debug.h"

#include "util.h"
#include "config.h"
#include "equation.h"
#include "lossfraction.h"


#define LOSS_V2
#undef LOSS_V3
/* #define QUIET */

int     ppl_cur;
int     ppl_hist[HIST_SIZE];
int     ts_hist[HIST_SIZE];       /* in ms */
int     seq_hist[HIST_SIZE];

u_int32 round;
int     num_loss;
int32   mseq_offs, high_offs;
int32   ppl_start, ppl_end;

int     weights;
int     const_weights;
float   weight[MAX_WEIGHTS];

int     ppl[MAX_WEIGHTS + 1];
/* float   mult[MAX_WEIGHTS + 1]; */
float   new_mult; /* deweighting factor */

double  (*lossfraction)();

void init_globals() {
        int i;
        for (i = 0; i < HIST_SIZE; ++i) {
                ppl_hist[i] = 0;
                ts_hist[i] = 0;
                seq_hist[i] = 0;
        }
        for (i = 0; i <= MAX_WEIGHTS; ++i) {
                ppl[i] = 0;
                /* mult[i] = 1.0; */
        }
        mseq_offs = 0;
        high_offs = -1;
        ppl_start = ppl_end = 0;
        ppl_cur = 0;
        weights = 8;
        round = 0;
        num_loss = 0;
        const_weights = 0;
        new_mult = 1.0;

#ifdef LOSS_V2
        lossfraction = lossfraction_v2;
#else
#ifdef LOSS_V3
        lossfraction = lossfraction_v3;
#else
        lossfraction = lossfraction_v1;
#endif
#endif
}


void set_weights() {
        int i;

        if (const_weights) {
                for (i = 0; i < weights; ++i)
                        weight[i] = 1.00;
        } else {
                double delta = 1.0 / ((weights + 1) / 2 + 1);
                for (i = 0; i < weights / 2; ++i)
                        weight[i] = 1.00;
                for (i = weights / 2; i < weights; ++i)
                        weight[i] = weight[i-1] - delta;
        }
        printf("weights: ");
        for (i = 0; i < weights; ++i)
                printf("%3.2f ", weight[i]);
        printf("\n");
}


void log_loss(int64 arrival) {
        int i;
        double l_fake;

        ppl[ppl_cur] += mseq_offs - high_offs - 1; /* account for missing packets */
        
        /* log params */
        ppl_hist[num_loss - 1] = ppl[ppl_cur];
        ts_hist[num_loss - 1] = arrival / MS_TO_US;
        seq_hist[num_loss - 1] = high_offs + 1;
                                        
        high_offs = mseq_offs; /* forget about packets missing inbetween */

#ifdef LOSS_V2
        /* shift and update mult array and reset new_mult */
        /* no longer necessary because of history faking
        for (i = weights; i > 0; --i) {
                mult[i] = max(mult[i - 1] * new_mult, MIN_MULT);
        }
        mult[0] = 1.0;
        */
        if (new_mult < 1.0) {
                /* fake history */
#ifndef QUIET
                printf("%f: history remodeled after deweighting\n",
                       (double) arrival / MICROSEC);
#endif
                l_fake = lossfraction();
                new_mult = 1.0; /* reset deweighting factor */
                for (i = 0; i <= MAX_WEIGHTS; ++i) {
                        if (ppl[i] > 0) /* only fake intervals that were already used */
                                ppl[i] = (int) (1.0 / l_fake + 0.5);
                }
        }
#endif

        ppl_cur = (ppl_cur + 1) % (weights + 1);
        ppl[ppl_cur] = 0;
}


int update_history(u_int32 mseq, int64 arrival, double rtt, u_int32 tzero, double b_act) {
        static u_int32 seq_offset = 0xFFFFFFFF;
        static int num_reordered = 0;
        static int fake_hist = 0;
        static int first_loss_ignored = 0;
        static int64 lastloss = 0;
        static u_int32 lastround = 0;
        static int32 seq_reordered[2];

        /* init seq_offset with first mseq received */
        if (seq_offset == 0xFFFFFFFF)
                seq_offset = mseq;
        /* ignore packets with sequence number lower than that of first packet (out of order) */
        if (mseq < seq_offset)
                return 1;
        mseq_offs = mseq - seq_offset;


        if (mseq_offs > high_offs) {
                if ((arrival - lastloss > rtt * MS_TO_US) && (round > lastround) && (mseq_offs > high_offs + 1)) {
                        /* check if missing packet(s) already arrived
                           and the reordered packets can now be inserted in the history */
                        if ((num_reordered > 0) && (seq_reordered[0] == high_offs + 1)) {
                                if ((num_reordered == 2) && (seq_reordered[1] == high_offs + 2)) {
                                        high_offs += 2;
                                        num_reordered = 0;
                                } else {
                                        seq_reordered[0] = seq_reordered[1];
                                        --num_reordered;
                                }
                        }
                        
                        /* high_offs might have changed, check again for missing packet */
                        if (mseq_offs > high_offs + 1) {
                                /* (emulate triple duplicate ACK) */
                                if (num_reordered < 2) { 
                                        if (num_reordered == 1) {
                                                if (seq_reordered[0] > mseq_offs) {
                                                        seq_reordered[1] = seq_reordered[0];
                                                        seq_reordered[0] = mseq_offs;
                                                } else {
                                                        seq_reordered[1] = mseq_offs;
                                                }
                                        } else {
                                                seq_reordered[0] = mseq_offs;
                                        }
                                        ++num_reordered;
                                        /*
                                          printf("%f: %i. packet out of order, seq %li (diff %li)\n",                                          (double) arrival / MICROSEC, num_reordered, mseq, mseq_offs - high_offs);
                                          */
                                } else {
                                        /* loss event */
                                        if((seq_reordered[0] - high_offs == 2) && (first_loss_ignored == 0)) {
                                                /* ignore first SINGLE loss */
#ifndef QUIET
                                                printf("%f: first loss event, seq %li IGNORED\n",
                                                       (double) arrival / MICROSEC, high_offs + seq_offset - 1);
#endif
                                                high_offs = mseq_offs; /* forget about missing packet */
                                                num_reordered = 0;
                                        } else {
                                                ++num_loss;
                                                if (num_loss >= HIST_SIZE) {
                                                        fprintf(stderr, "loss history buffer overflow\n");
                                                        return 0;
                                                }
#ifndef QUIET
                                                printf("%f: loss event, seq %li\n",
                                                       (double) arrival / MICROSEC, high_offs + seq_offset - 1);
#endif
                                                if (fake_hist == 0) {
                                                        /* fake history (once) -> half send rate if that results in
                                                           a smaller initial loss rate, otherwise it was probably a 
                                                           random lost packet */
                                                        fake_history((int) (1.0 / lossfraction_eq(b_act / 2.0, rtt, tzero)));
                                                        fake_hist = 1;
#ifndef QUIET
                                                        printf("\t fake_ppl = %i\n", ppl[0]);
#endif
                                                }
                                                log_loss(arrival);
                                                num_reordered = 0;
                                                lastloss = arrival;
                                                lastround = round;
                                        }
                                        first_loss_ignored = 1;
                                }
                        }
                } else {
                        /* packet in order or less than one RTT from previous loss */
                        high_offs = mseq_offs;
                }
                ++ppl[ppl_cur]; /* add packet to last interval */
        } else {
                /* ignore reordered packet (after TD ACK) or duplicate packet */
#ifndef QUIET
                if (mseq == high_offs + seq_offset)
                        printf("%f: duplicate packet ignored (seq %li)\n",
                               (double) arrival / MICROSEC, mseq);
                else
                        printf("%f: reordered packet ignored (seq %li, highest seq %li)\n",
                               (double) arrival / MICROSEC, mseq, high_offs + seq_offset);
#endif
        }
        return 1;
}


u_int32 calc_tzero(double rtt) {
	static int t_srtt_ = 8;
	static int t_rttvar_ = 0;
	double t_rtxcur_;
	int t_rtt_;

        t_rtt_ = (int) (rtt / MILLISEC / tcp_tick_ + 0.5);
        if (t_rtt_ < 1)
                t_rtt_ = 1;

	/*
          srtt has 3 bits to the right of the binary point
          rttvar has 2
          */
        
        if (t_srtt_ != 0) {
                register short delta;
                delta = t_rtt_ - (t_srtt_ >> T_SRTT_BITS);  
                if ((t_srtt_ += delta) <= 0)  
                        t_srtt_ = 1;
                if (delta < 0)
                        delta = -delta;
                delta -= (t_rttvar_ >> T_RTTVAR_BITS);
                if ((t_rttvar_ += delta) <= 0)  
                        t_rttvar_ = 1;
        } else {
                t_srtt_ = t_rtt_ << T_SRTT_BITS;    
                t_rttvar_ = t_rtt_ << (T_RTTVAR_BITS-1);  
        }
	
	/*
          Current retransmit value is (unscaled) smoothed round trip estimate
          plus 4 times (unscaled) rttvar.
          */

        t_rtxcur_ = (((t_rttvar_ << (2 + (T_SRTT_BITS - T_RTTVAR_BITS))) +
                      t_srtt_)  >> T_SRTT_BITS ) * tcp_tick_;
        if (t_rtxcur_ < 2 * tcp_tick_)
                t_rtxcur_ = 2 * tcp_tick_;
        return t_rtxcur_ * MILLISEC;
}


int main(int argc, char *argv[]) {
        char *addr = NULL;
        int  port = -1;
        int  ttl = 16;

        char send_buffer[REP_SIZE];
        char recv_buffer[PKT_SIZE];
        struct pkt_dtg *packet = (struct pkt_dtg *)recv_buffer;
        struct rep_dtg *rep    = (struct rep_dtg *)send_buffer;

        socket_udp *send_socket;
        socket_udp *recv_socket;

        int64   sim_start = 0, sim_end = 0;
        int64   arrival = 0, last_report = 0, last_flush = 0;
        /* int64   last_brep = 0; */
        /* int64   mni_ts; */                              /* time when M_NO_INCREASE mode was entered */

        u_int32 rtt_cur = 0;
        u_int32 rtt_cur_old = 0;                     /* rtt_cur of previous packet */
        u_int32 tzero = 0;
        double  rtt_ema = 0.0;
        double  l_act = 0.0;
        double  b_act = 0, b_exp = 0, b_rep = 0;
        /* double  b_exp_max = 0; */

        int     duration = 0;                        /* quit after 'duration' seconds */
        int     datalen = 0, data_per_report = 0;
        struct  timeval max_wait;
        int     nready;
        u_int32 mode = M_INITRTT;
        /* int     num_rttinc = 0; */                      /* number of consecutive incs in RTT */
        int     packets = 0;
        u_int32 mseq;
        int     log_nr = 1;
        char    log_name[100];
        FILE *h_file;

        int32 i;

        init_globals();
        
        if (argc < 3) {
                fprintf(stderr, "usage: %s address port [ttl] [-t duration] [-n lognr] [-w num. weights]\n", argv[0]);
                exit(1);
        } else {
                addr = argv[1];
                port = atoi(argv[2]);
                for (i = 3; i < argc; ++i) {
                        if (argv[i][0] != '-') {
                                ttl = atoi(argv[i]);
                        } else {
                                switch (argv[i][1]) {
                                case 'n' : if (argc > i + 1) log_nr = atoi(argv[++i]); break;
                                case 't' : if (argc > i + 1) duration = atoi(argv[++i]); break;
                                case 'w' : if (argc > i + 1) weights = atoi(argv[++i]); break;
                                case 'c' : const_weights = 1; break;
                                default : fprintf(stderr, "unknown option: %s\n", argv[i]); exit(1); break;
                                }
                        }
                }
        }
        set_weights();

#ifndef QUIET
        printf("started ");
        for (i = 0; i < argc; ++i)
                printf("%s ", argv[i]);
        printf("\n");
#endif

        udp_fd_zero();
        send_socket = udp_init(addr, port+1, ttl);
        recv_socket = udp_init(addr, port  , ttl);
        if (send_socket == NULL || recv_socket == NULL) {
                perror("open socket");
                exit(1);
        }
        
        sprintf(log_name, "all%i.log", log_nr);
        h_file = fopen(log_name, "w");
        if (h_file == NULL) {
                perror("open log file");
                exit(1);
        }

        init_time();
        sim_start = get_time() + 1; /* init sim_start */
        if (duration > 0)
                sim_end = sim_start + duration * MICROSEC;

        while (sim_end == 0 || get_time() < sim_end) {
                udp_fd_set(recv_socket);
                max_wait.tv_sec = (sim_end - get_time()) / MICROSEC + 1;
                max_wait.tv_usec = 0;
                if ((nready = udp_select((duration > 0) ? &max_wait : NULL)) <= 0) {
                        if (nready == 0)
                                break;
                        if (errno == EINTR)
                                continue;
                        else
                                perror("select_error");
                }
                if (udp_fd_isset(recv_socket)) {
                        datalen = udp_recv(recv_socket, recv_buffer, PKT_SIZE);
                        data_per_report += datalen;

                        ++packets;
                        arrival = get_time();
                        mseq    = ntohl(packet->mseq);
                        b_act   = ntohl(packet->bitrate);
                        rtt_cur = ntohl(packet->rtt);

                        /* rtt_ema = ntohl(packet->rtt); */
                        /* if (mode != M_NO_INCREASE) */
                        mode = ntohl(packet->mode); /* wait until RTT decreases -> M_NORMAL */

                        if (rtt_ema > 0.0)
                                rtt_ema = (1 - RTT_DECAY) * rtt_ema + RTT_DECAY * rtt_cur;
                        else
                                rtt_ema = rtt_cur;

#ifdef NOTDEF
                        /* analyzation of inc RTTs no longer needed with inter packet space modulation */
                        if (rtt_cur > rtt_cur_old && mode == M_NORMAL) {
                                ++num_rttinc;
                                if (num_rttinc >= BUF_INDICATOR) {
                                        mode = M_NO_INCREASE;
                                        mni_ts = arrival;
                                        b_exp_max = b_exp;
                                        /* printf("%f: receiver in M_NO_INCREASE mode\n", (double) arrival / MICROSEC); */
                                }
                        } else if (rtt_cur < rtt_cur_old) {
                                if (mode == M_NO_INCREASE) {
                                        mode = M_NORMAL;
                                        /* printf("%f: receiver back to M_NORMAL mode\n", (double) arrival / MICROSEC); */
                                }
                                num_rttinc = 0;
                        }
                        /* timeout M_NO_INCREASE after 10 RTTs, XXX not very elegant */
                        if (mode == M_NO_INCREASE && (arrival - mni_ts) / MS_TO_US > 10 * rtt_ema) {
                                mode = M_NORMAL;
                                num_rttinc = 0;
                                /* printf("%f: receiver back to M_NORMAL mode (timeout)\n", (double) arrival / MICROSEC); */
                        }
#endif

                        /* tzero = calc_tzero(rtt_ema); unused !! */
                        tzero = max(4 * max(rtt_ema, rtt_cur), 500);
                        /* tzero = 4 * rtt_ema; */
                        rtt_cur_old = rtt_cur;
                        
                        /* adjust round number only if sender has responded to the first loss event
                         * and kicked out of SLOWSTART
                         */
                        if (!(mode == M_SLOWSTART && num_loss > 0))
                                round = ntohl(packet->round);

                        if (update_history(mseq, arrival, rtt_cur, tzero, b_act) == 0)
                                break; /* -> error */

                        /*
                         * send REPORTS_PER_RTT reports per rtt (when at least one packet was received
                         * since last report), send report for every packet in M_INITRTT mode and when we 
                         * encountered loss and are still in M_SLOWSTART
                         */
                        if (((arrival - last_report) / MS_TO_US >= rtt_ema / REPORTS_PER_RTT) ||
                            (mode == M_INITRTT) || (mode == M_SLOWSTART && num_loss > 0)) {
                                int ret;

                                /* update reported bitrate every RTT */
                                /*if ((arrival - last_brep) / MS_TO_US >= rtt_ema) {
                                        if (b_rep == 0.0) {
                                                b_rep = data_per_report / 
                                                        ((double) (arrival - last_brep) / MICROSEC);
                                        } else {
                                                b_rep =  (1 - BREP_DECAY) * b_rep + BREP_DECAY *
                                                        data_per_report / ((double) (arrival - last_brep) / MICROSEC);
                                        }
                                        data_per_report = 0;
                                        last_brep = arrival;
                                }*/

                                /* update reported bitrate every report */
                                b_rep = data_per_report / ((double) (arrival - last_report) / MICROSEC);
                                data_per_report = 0;
                                 
                                /* flush every 500 ms */
                                if (arrival - last_flush >= 500000) {
                                        fflush(h_file);
                                        last_flush = arrival;
                                }

                                /* calculate expected bitrate */
                                if (num_loss > 0) {
                                        l_act = lossfraction();
                                        b_exp = bitrate_eq(l_act, rtt_ema, tzero);
                                } else {
                                        l_act = 0.0;
                                        b_exp = 0.0;
                                }

#ifdef NOTDEF
                                /* not used, see above */
                                /* only increase above b_exp_max if not in M_NO_INCREASE mode */
                                if (mode == M_NO_INCREASE && b_exp > b_exp_max) {
                                        b_exp = b_exp_max;
                                        if (b_exp - b_exp_max > 10)
                                                printf("%f: M_NO_INCREASE mode, bitrate inc. prevented (%2.1f / %2.1f)\n",
                                                       (double) arrival / MICROSEC, b_exp, b_exp_max);
                                }
#endif

                                /* report lossrate etc. */
                                rep->mseq     = packet->mseq;
                                rep->dseq     = packet->dseq;
                                rep->ts       = packet->ts;
                                rep->b_rep    = htonl((u_int32) b_rep);
                                rep->b_exp    = htonl((u_int32) b_exp);
                                while ((ret = udp_send(send_socket, (char *) rep, REP_SIZE)) < 0) {
                                        fprintf(stderr, "%3.2f ", (float) get_time() / MICROSEC);
                                        perror("udp_send");
                                }

                                last_report = arrival;
                                packets = 0;
                        }
                        /* log relevant variables */
                        if (mode != M_INITRTT) {
                                /* 1 time, 2 seqno, 3 round, 4 loss, 5 bact, 6 bexp, 7 brep, 8 rtt_ema, 9 rtt_cur, 10 tzero, ... */
                                fprintf(h_file, "%f %li %lu %f %1.0f %1.0f %1.0f %f %lu %lu",
                                        (float) (arrival - sim_start) / MICROSEC, mseq, round, l_act, b_act,
                                        b_exp, b_rep, rtt_ema, rtt_cur, tzero);
                                /* 11 ppl_start, 12 ppl_end */
                                fprintf(h_file, " %li %li", ppl_start, ppl_end);
                                for (i = 0; i <= weights; ++i)
                                        fprintf(h_file, " %i", ppl[(ppl_cur - i + weights + 1) % (weights + 1)]);
                                /* fprintf(h_file, "\n"); */
                                fprintf(h_file, " %f\n", new_mult);
                        }
                }
        }
        printf("%f: receiver finished, writing history files\n", (float) get_time() / MICROSEC);
        fclose(h_file);

        sprintf(log_name, "losshist%i.log", log_nr);
        h_file = fopen(log_name, "w");
        for (i = 0; i < num_loss; ++i)
                fprintf(h_file, "%f %i %i\n", (float) ts_hist[i] / MILLISEC, ppl_hist[i], seq_hist[i]);
        fclose(h_file);

        sprintf(log_name, "weights%i.log", log_nr);
        h_file = fopen(log_name, "w");
        for (i = 0; i < weights; ++i)
                fprintf(h_file, "%f ", weight[i]);
        fprintf(h_file, "\n");
        fclose(h_file);
        return 0;
}
