#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <strings.h>
#include <math.h>

#include "config_unix.h"
#include "config_win32.h"
#include "debug.h"

#include "util.h"
#include "config.h"

int main(int argc, char **argv) {
        int sockfd, recvsockfd;
        int recvlen, datalen;
        struct sockaddr_in recv_addr;
        char data_buf[PKT_SIZE];
        int tcpbuf_size = 200000;

        int64 sim_end = 0;
        int duration = 0;
        int sync = 0;
        int log = 0;
        int log_nr = 0;
        short port;
        int i;

        FILE *h_file;
        char log_name[100];

        if (argc < 2) {
                fprintf(stderr, "usage: %s port [-t duration] [-s] [-l]\n", argv[0]);
                exit(1);
        } else {
                port = atoi(argv[1]);
                for (i = 2; i < argc; ++i) {
                        if (argv[i][0] == '-') {
                                switch (argv[i][1]) {
                                case 't' : if (argc > i + 1) duration =  atoi(argv[++i]); break;
                                case 'n' : if (argc > i + 1) log_nr =  atoi(argv[++i]); break;
                                case 's' : sync = 1; break;
                                case 'l' : log = 1; break;
                                default : break;
                                }
                        }
                }
        }

        if (sync == 1)
                log = 0;

        printf("started ");
        for (i = 0; i < argc; ++i)
                printf("%s ", argv[i]);
        printf("\n");
        
        init_time();
        if (duration > 0)
                sim_end = get_time() + duration * MICROSEC; /* init sim_end */

        if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
                perror ("server: can't open stream socket");
                exit(1);
        }

        bzero((char*)&recv_addr, sizeof(recv_addr));
        recv_addr.sin_family = AF_INET ;
        recv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        recv_addr.sin_port = htons(port);

        setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, (void *) &tcpbuf_size, sizeof(int));
        setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, (void *) &tcpbuf_size, sizeof(int));

        if (bind(sockfd, (struct sockaddr *) &recv_addr, sizeof(recv_addr)) < 0) {
                perror ("server: bind failed");
                exit(1);
        }

        listen(sockfd, 5);
	recvlen = sizeof(recv_addr);
	recvsockfd = accept(sockfd, (struct sockaddr *) &recv_addr, &recvlen);
        
        if (log == 1) {
                sprintf(log_name, "tcp_br%i.log", log_nr);
                h_file = fopen(log_name, "w");
        }

        while (duration == 0 || get_time() < sim_end) {
                datalen = read(recvsockfd, data_buf, PKT_SIZE);
                if (datalen < 0) {
                        perror("tcp: read error");
                        exit(1);
                }
                if (sync == 1) break;
                if (log == 1)
                        fprintf(h_file, "%f %i\n", (double) get_time() / MICROSEC, datalen);
        }
        if (log == 1)
                fclose(h_file);
	close(recvsockfd);
        printf("%f: tcpreceiver finished\n", (double) get_time() / MICROSEC);
        return 0;
}
